/*
 * Copyright (c) 2010-2012 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.vew.autoexport.AutoExport
 *
 * de.inovat.vew.autoexport.AutoExport is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.vew.autoexport.AutoExport is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.vew.autoexport.AutoExport; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.inovat.vew.autoexport;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import static java.util.Calendar.HOUR_OF_DAY;
import static java.util.GregorianCalendar.MILLISECOND;
import static java.util.GregorianCalendar.MINUTE;
import static java.util.GregorianCalendar.SECOND;

//~ JDK IMPORTE ===============================================================

//~ KLASSEN ===================================================================

/**
 * Klasse, die den automatischen Export der BASt-Band Erzeugung fr LZZS fr die Formate V2004, V2007 und AXL
 * durchfhrt.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kni (HCK)
 * @version $Revision: 780 $ / $Date: 2012-10-31 11:44:41 +0100 (Mi, 31 Okt 2012) $ / ($Author: HCK $)
 */
public class AutoExport implements StandardApplication, ClientSenderInterface {

    /** WarteZeit in ms, bis die Auftragsliste erneut geprft wird, nachdem keinerlei Auftrge mehr vorlagen * */
    private static final long WARTEZEIT_AUF_NAECHSTEN_AUFTRAG = 60 * 1000;  // 1 Minute

    /** DebugLogger fuer Debug-Ausgaben. */
    private static Debug debug;

    //~ FELDER ================================================================

    private long _bisZeitPunktAXL;
    private long _bisZeitPunktLZZS;

    /** Datenverteilerverbindung */
    private ClientDavInterface _dav;

    /** Legt fest, ob die FTP-Verbindung im PassivModus eingebaut werden soll. */
    private String  _ftpPassivModus;
    private String  _ftpPassivModusAufrufParameter;
    private String  _ftpServerBenutzer;
    private String  _ftpServerBenutzerAufrufParameter;
    private String  _ftpServerPasswort;
    private String  _ftpServerPasswortAufrufParameter;
    private String  _ftpServerURL;
    private String  _ftpServerURLAufrufParameter;
    private boolean _isAenderungDatenstruktur;
    private boolean _isAenderungDatenstrukturAufrufParameter;
    private boolean _isBastBandUeberschreiben;
    private boolean _isBastBandUeberschreibenAufrufParameter;

    /** Die Konfiguration des DaV. */
    private DataModel          _konfiguration;
    private String             _listePidsAufrufParameter;
    private String             _minimaleWarteZeitInStundenAufrufParameterAXL;
    private String             _minimaleWarteZeitInStundenAufrufParameterLZZS;
    private int                _minimaleWarteZeitInStundenAXL;
    private int                _minimaleWarteZeitInStundenLZZS;
    private String             _pidVerwaltungsObjektBastExportSteuerung;
    private SystemObject       _verwaltungsObjektBastExportSteuerung;
    private long               _vonZeitPunktAufrufParameterAXL;
    private long               _vonZeitPunktAufrufParameterLZZS;
    private long               _vonZeitPunktAXL;
    private long               _vonZeitPunktLZZS;
    private List<String>       _listePidsLzzs               = new ArrayList<String>();
    private List<SystemObject> _listeLzzsSystemObjekte      = new ArrayList<SystemObject>();
    private List<String>       _listePidsLzzsNichtVorhanden = new ArrayList<String>();

    //~ METHODEN ==============================================================

    /**
     * Main-Methode der SWE AutoExport.
     *
     * @param args Aufrufargumente der Applikation.
     */
    public static void main(String[] args) {
        AutoExport autoExport = new AutoExport();

        try {
            StandardApplicationRunner.run(autoExport, args);

            // Debug erst hier initialisieren, da erst ab hier verfgbar !!
            debug = Debug.getLogger();
            autoExport.auswertungAufrufParameterPidVerwaltungsObjekt();
            autoExport.auswertungAufrufParameterListePids();
            autoExport.auswertungAufrufParameterVonZeitPunktLZZS();
            autoExport.auswertungAufrufParameterVonZeitPunktAXL();
            autoExport.auswertungAufrufParameterBAStUndFTP();
            autoExport.initialisierung();
            debug.config(autoExport.ausgabeKonfigurationsZusammenfassung());
        }
        catch (RuntimeException e) {
            e.printStackTrace();
            System.exit(-1);
        }

        autoExport.run();
    }

    /**
     * Gibt Informationen zur Konfiguration aus. Ist nur zu Anlaysezwecken bzw. zur zusammenfassenden Ausgabe der
     * Konfigurationsphase dieser Applikatione interessant.
     *
     * @return Informationen zur Konfiguration (Aufbau der Ausgabe kann sich ndern).
     */
    private String ausgabeKonfigurationsZusammenfassung() {
        StringBuilder sb = new StringBuilder();

        // Listen der SystemObjekte sortieren...
        Collections.sort(_listeLzzsSystemObjekte, new Comparator<SystemObject>() {
            public int compare(SystemObject o1, SystemObject o2) {
                return o1.getPid().compareTo(o2.getPid());
            }
        });
        Collections.sort(_listePidsLzzsNichtVorhanden);
        sb.append(String.format("%n%n"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("START: Konfigurationszusammenfassung fr AuoExportBASt"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -pidVerwaltungsObjekt=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _pidVerwaltungsObjektBastExportSteuerung));

        if (_verwaltungsObjektBastExportSteuerung != null) {
            sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _verwaltungsObjektBastExportSteuerung.getPid()));
        } else {
            sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _verwaltungsObjektBastExportSteuerung));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Im Aufrufparameter angegebene PIDs, zu denen keine SystemObjekte existieren"));
        sb.append(String.format("%n====================================================================================%n"));

        int i = 0;

        for (String pid : _listePidsLzzsNichtVorhanden) {
            sb.append(String.format("%04d: %s%n", ++i, pid));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Liste der PIDs der verwendeten Systemobjekte"));
        sb.append(String.format("%n====================================================================================%n"));

        int j = 0;

        for (SystemObject so : _listeLzzsSystemObjekte) {
            sb.append(String.format("%04d: %s%n", ++j, so.getPidOrNameOrId()));
        }

        SimpleDateFormat datumsFormatierer = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss,SSS");

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -vonZeitPunktLZZS=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunktAufrufParameterLZZS))));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunktLZZS))));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -vonZeitPunktAXL=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunktAufrufParameterAXL))));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunktAXL))));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -minimaleWarteZeitInStundenLZZS=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _minimaleWarteZeitInStundenAufrufParameterLZZS));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _minimaleWarteZeitInStundenLZZS));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -minimaleWarteZeitInStundenAXL=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _minimaleWarteZeitInStundenAufrufParameterAXL));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _minimaleWarteZeitInStundenAXL));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -bastBandUeberschreiben=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _isBastBandUeberschreibenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _isBastBandUeberschreiben));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -aenderungDatenstruktur=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _isAenderungDatenstrukturAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _isAenderungDatenstruktur));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -ftpServerBenutzer=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _ftpServerBenutzerAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _ftpServerBenutzer));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -ftpServerPasswort=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _ftpServerPasswortAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _ftpServerPasswort));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -ftpServerURL=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _ftpServerURLAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _ftpServerURL));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -ftpPassivModus=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _ftpPassivModusAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _ftpPassivModus));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Auftragsliste initial..."));
        sb.append(String.format("%n====================================================================================%n"));

        int k = 0;

        for (Auftrag auftrag : AuftragsListe.getInstanz().holeKopieAuftragsListe()) {
            sb.append(String.format("%04d: %s%n", ++k, auftrag.toString()));
        }

        sb.append(String.format("%n%n"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("ENDE: Konfigurationszusammenfassung fr AuoExportBASt"));
        sb.append(String.format("%n####################################################################################%n"));

        return sb.toString();
    }

    private void auswertungAufrufParameterBAStUndFTP() {

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -bastBandUeberschreiben
        // ------------------------------------------------------------------------------------------------------
        _isBastBandUeberschreiben = _isBastBandUeberschreibenAufrufParameter;

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -aenderungDatenstruktur
        // ------------------------------------------------------------------------------------------------------
        _isAenderungDatenstruktur = _isAenderungDatenstrukturAufrufParameter;

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -ftpServer
        // ------------------------------------------------------------------------------------------------------
        _ftpServerBenutzer = _ftpServerBenutzerAufrufParameter;
        _ftpServerPasswort = _ftpServerPasswortAufrufParameter;
        _ftpServerURL      = _ftpServerURLAufrufParameter;
        _ftpPassivModus    = _ftpPassivModusAufrufParameter;
    }

    /**
     * Wertet die Aufrufparameter
     * <p/>
     * -listePids
     * <p/>
     * wie folgt aus:
     * <p/>
     * Es werden alle Systemobjekte vom Typ <tt>typ.langZeitZhlstelle</tt> ermittelt und dabei diejenigen fr die
     * weiteren Berechnungen bernommen, deren Pid in der Liste der Pids aufgefhrt ist (bzw. alle, falls Aufruf mit
     * <tt>-listePids=alle</tt> erfolgt). Wird kein passendes Objekt gefunden, wird das Programm mit einer Fehlermeldung
     * abgebrochen.
     */
    private void auswertungAufrufParameterListePids() {
        try {

            // Pids der zu bercksichtigenden Objekte ermitteln
            _listePidsLzzs = Arrays.asList(_listePidsAufrufParameter.split("(:|,|;)"));

            // Alle mglichen SystemObjekte ermittlen...
            List<SystemObject> listeAlleLzzsSystemObjekte = _konfiguration.getType("typ.langZeitZhlstelle").getElements();

            // ... und jetzt Liste mit zu verwendenden aufbauen (falls nicht "alle" verwendet werden sollen) ...
            boolean isAlle = ((_listePidsLzzs.size() == 1) && _listePidsLzzs.get(0).equals("alle"));

            for (SystemObject so : listeAlleLzzsSystemObjekte) {
                if (isAlle || _listePidsLzzs.contains(so.getPid())) {
                    _listeLzzsSystemObjekte.add(so);
                }
            }

            // .. und fr Konfigurationszusammenfassung Liste der nicht verwendeten PID's aus dem Aufrufparameter ermitteln
            List<String> listePidsVerwendet = new ArrayList<String>();

            for (SystemObject so : _listeLzzsSystemObjekte) {
                listePidsVerwendet.add(so.getPid());
            }

            for (String pid : _listePidsLzzs) {
                if (!listePidsVerwendet.contains(pid)) {
                    _listePidsLzzsNichtVorhanden.add(pid);
                }
            }
        }
        catch (Exception e) {
            debug.error("Es konnten keine Objekte vom typ.langZeitZhlstelle ermittelt werden!. Programm wird beendet.", e);

            throw new RuntimeException();
        }

        if (_listeLzzsSystemObjekte.size() == 0) {
            debug.error("Es konnten keine Objekte vom typ.langZeitZhlstelle ermittelt werden!. Programm wird beendet.");

            throw new RuntimeException();
        }
    }

    /**
     * Wertet den Aufrufparameter
     * <p/>
     * -pidVerwaltungsObjekt
     * <p/>
     * wie folgt aus:
     * <p/>
     * Ist der Wert <code>automatisch</code>, so wird das Objekt mit dem Typ <code>typ.bastExportImportSteuerung</code>
     * und der Pid <code>bastBand.niDauz</code> verwendet, sonst das Objekt mit der im Aufrufparameter bergebenen
     * Pid..
     */
    private void auswertungAufrufParameterPidVerwaltungsObjekt() {
        try {
            if (_pidVerwaltungsObjektBastExportSteuerung.equals("automatisch")) {
                _verwaltungsObjektBastExportSteuerung = _konfiguration.getObject("bastBand.niDauz");
            } else {
                _verwaltungsObjektBastExportSteuerung = _konfiguration.getObject(_pidVerwaltungsObjektBastExportSteuerung);
            }

            if (_verwaltungsObjektBastExportSteuerung != null) {
                if (!_verwaltungsObjektBastExportSteuerung.isOfType("typ.bastExportImportSteuerung")) {
                    debug.error(String.format("VerwaltungsObjekt des AutoExports der BASt-Band-Steuerung [-pidVerwaltungsObjektBastExportSteuerung=%s] ist nicht vom Typ [typ.bastExportImportSteuerung] sondern vom Typ [%s].%nProgramm wird beendet.", _pidVerwaltungsObjektBastExportSteuerung, _verwaltungsObjektBastExportSteuerung.getType().getPid()));
                    System.exit(-1);
                }
            } else {
                debug.error(String.format("VerwaltungsObjekt des AutoExports der BASt-Band-Steuerung [-pidVerwaltungsObjektBastExportSteuerung=%s] konnte nicht ermittelt werden.%nProgramm wird beendet.", _pidVerwaltungsObjektBastExportSteuerung));
                System.exit(-1);
            }
        }
        catch (Exception e) {
            debug.error(String.format("VerwaltungsObjekt des AutoExports der BASt-Band-Steuerung [-pidVerwaltungsObjektBastExportSteuerung=%s] konnte nicht ermittelt werden.%nProgramm wird beendet.", _pidVerwaltungsObjektBastExportSteuerung));
            System.exit(-1);
        }
    }

    /**
     * Wertet den Aufrufparameter
     * <p/>
     * -vonZeitPunktAXL
     * <p/>
     * wie folgt aus:
     * <p/>
     * <tt>-vonZeitPunktAXL</tt>: Wird auf gltigen Intervallanfangszeitpunkt des laufenden Tages gesetzt.
     * <p/>
     * Beispiel: Ist <tt>-vonZeitPunktAXL=04.11.2010 11:51:01,956</tt>,  so wird der Zeitpunkt auf <tt>04.11.2010
     * 00:00:00,000</tt> gesetzt.
     * <p/>
     * Der Endzeitpunkt des Intervalls wird auf gltigen Intervallanfangszeitpunkt des nchsten Tages (dahinterliegender
     * Intervallbeginn) gesetzt, im vorliegenden Beispiel also auf den 05.11.2010 00:00:00,000.
     */
    private void auswertungAufrufParameterVonZeitPunktAXL() {

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunktAXL
        // ------------------------------------------------------------------------------------------------------
        _vonZeitPunktAXL = ermittleIntervallBeginnAXL(_vonZeitPunktAufrufParameterAXL);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -bisZeitPunktAXL
        // ------------------------------------------------------------------------------------------------------
        _bisZeitPunktAXL = ermittleIntervallEndeAXL(_vonZeitPunktAufrufParameterAXL);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInStundenAXL
        // ------------------------------------------------------------------------------------------------------
        try {
            _minimaleWarteZeitInStundenAXL = new Integer(_minimaleWarteZeitInStundenAufrufParameterAXL);

            if (_minimaleWarteZeitInStundenAXL <= 0) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _minimaleWarteZeitInStundenAXL = 1;
            debug.warning(String.format("Aufrufparameter [-minimaleWarteZeitInStundenAXL=%s] ist ungltig. Wert wurde fr weiter Berechnung auf [-minimaleWarteZeitInStundenAXL=%d] gesetzt.", _minimaleWarteZeitInStundenAufrufParameterAXL, _minimaleWarteZeitInStundenAXL));
        }
    }

    /**
     * Wertet den Aufrufparameter
     * <p/>
     * -vonZeitPunktLZZS
     * <p/>
     * wie folgt aus:
     * <p/>
     * <tt>-vonZeitPunktLZZS</tt>: Wird auf gltigen Intervallanfangszeitpunkt des laufenden Monats gesetzt.
     * <p/>
     * Beispiel: Ist <tt>-vonZeitPunktLZZS=04.11.2010 11:51:01,956</tt>,  so wird der Zeitpunkt auf <tt>01.11.2010
     * 00:00:00,000</tt> gesetzt.
     * <p/>
     * Der Endzeitpunkt des Intervalls wird auf gltigen Intervallanfangszeitpunkt des nchsten Monats
     * (dahinterliegender Intervallbeginn) gesetzt, im vorliegenden Beispiel also auf den 01.12.2010 00:00:00,000.
     */
    private void auswertungAufrufParameterVonZeitPunktLZZS() {

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunktLZZS
        // ------------------------------------------------------------------------------------------------------
        _vonZeitPunktLZZS = ermittleIntervallBeginnLZZS(_vonZeitPunktAufrufParameterLZZS);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -bisZeitPunktLZZS
        // ------------------------------------------------------------------------------------------------------
        _bisZeitPunktLZZS = ermittleIntervallEndeLZZS(_vonZeitPunktAufrufParameterLZZS);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInStundenLZZS
        // ------------------------------------------------------------------------------------------------------
        try {
            _minimaleWarteZeitInStundenLZZS = new Integer(_minimaleWarteZeitInStundenAufrufParameterLZZS);

            if (_minimaleWarteZeitInStundenLZZS <= 0) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _minimaleWarteZeitInStundenLZZS = 1;
            debug.warning(String.format("Aufrufparameter [-minimaleWarteZeitInStundenLZZS=%s] ist ungltig. Wert wurde fr weiter Berechnung auf [-minimaleWarteZeitInStundenLZZS=%d] gesetzt.", _minimaleWarteZeitInStundenAufrufParameterLZZS, _minimaleWarteZeitInStundenLZZS));
        }
    }

    /**
     * Sendesteuerung des Datenverteilers an die Applikation. Diese Methode muss von der Applikation implementiert
     * werden, um den Versand von Daten zu starten bzw. anzuhalten. Der Datenverteiler signalisiert damit einer Quelle
     * oder einem Sender dass mindestens ein Abnehmer bzw. kein Abnehmer mehr fr die zuvor angemeldeten Daten vorhanden
     * ist. Die Quelle wird damit aufgefordert den Versand von Daten zu starten bzw. zu stoppen.
     *
     * @param object          Das in der zugehrigen Sendeanmeldung angegebene Objekt, auf das sich die Sendesteuerung
     *                        bezieht.
     * @param dataDescription Beschreibende Informationen zu den angemeldeten Daten auf die sich die Sendesteuerung
     *                        bezieht.
     * @param state           Status der Sendesteuerung. Kann einen der Werte <code>START_SENDING</code>,
     *                        <code>STOP_SENDING</code>, <code>STOP_SENDING_NO_RIGHTS</code>,
     *                        <code>STOP_SENDING_NOT_A_VALID_SUBSCRIPTION</code> enthalten.
     *
     * @see #START_SENDING
     * @see #STOP_SENDING
     * @see #STOP_SENDING_NO_RIGHTS
     * @see #STOP_SENDING_NOT_A_VALID_SUBSCRIPTION
     */
    @Override
    public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {

        // tut nicht, da Sendesteuerung nicht untersttzt wird
    }

    /**
     * Ermittelt den IntervallBeginn ausgehend von einem Zeitpunk und der Intervalldauer ein Tag
     *
     * @param spaetesterIntervallBeginn Anfangszeitpunkt in ms seit 1.1.1970, ab dem sptestens die Daten berechnet
     *                                  werden sollen. Dieser Wert wird dann noch soweit in die Vergangenheit
     *                                  verschoben, dass genau ein gltiger Intervallbeginn (Tagesbeginn) ermittelt
     *                                  wird.
     *
     * @return Intervallbeginn (GENAU!!), der vor dem Zeitpunkt <tt>vonZeitPunkt</tt> liegt
     */
    private long ermittleIntervallBeginnAXL(long spaetesterIntervallBeginn) {
        Calendar vonZeitPunkt = new GregorianCalendar();

        // vonZeitPunkt darf sptestens sein...
        vonZeitPunkt.setTimeInMillis(spaetesterIntervallBeginn);

        // ... und muss jetzt auf gltigen Intervallzeitpunkt DAVOR (Tagesanfang) abgebildet werden...
        vonZeitPunkt.set(MILLISECOND, 0);
        vonZeitPunkt.set(SECOND, 0);
        vonZeitPunkt.set(MINUTE, 0);
        vonZeitPunkt.set(HOUR_OF_DAY, 0);

        return vonZeitPunkt.getTimeInMillis();
    }

    /**
     * Ermittelt den IntervallBeginn ausgehend von einem Zeitpunk und der Intervalldauer ein Monat
     *
     * @param spaetesterIntervallBeginn Anfangszeitpunkt in ms seit 1.1.1970, ab dem sptestens die Daten berechnet
     *                                  werden sollen. Dieser Wert wird dann noch soweit in die Vergangenheit
     *                                  verschoben, dass genau ein gltiger Intervallbeginn (Monatsbeginn) ermittelt
     *                                  wird.
     *
     * @return Intervallbeginn (GENAU!!), der vor dem Zeitpunkt <tt>vonZeitPunkt</tt> liegt
     */
    private long ermittleIntervallBeginnLZZS(long spaetesterIntervallBeginn) {
        Calendar vonZeitPunkt = new GregorianCalendar();

        // vonZeitPunkt darf sptestens sein...
        vonZeitPunkt.setTimeInMillis(spaetesterIntervallBeginn);

        // ... und muss jetzt auf gltigen Intervallzeitpunkt DAVOR (Monatsanfang) abgebildet werden...
        vonZeitPunkt.set(MILLISECOND, 0);
        vonZeitPunkt.set(SECOND, 0);
        vonZeitPunkt.set(MINUTE, 0);
        vonZeitPunkt.set(HOUR_OF_DAY, 0);
        vonZeitPunkt.set(Calendar.DAY_OF_MONTH, 1);

        return vonZeitPunkt.getTimeInMillis();
    }

    /**
     * Ermittelt das IntervallEnde ausgehend von einem Zeitpunkt und der Intervalldauer Tag
     *
     * @param fruehestesIntervallEnde Endzeitpunkt, bis zu dem mindestens gerechnet werden muss.
     *
     * @return Intervallende (AUSSCHLIESSLICH!!), also der Intervallbeginn des nchsten Intervalls
     */
    private long ermittleIntervallEndeAXL(long fruehestesIntervallEnde) {
        Calendar bisZeitPunkt = new GregorianCalendar();

        bisZeitPunkt.setTimeInMillis(ermittleIntervallBeginnAXL(fruehestesIntervallEnde));

        // Zum Intervallanfang 1 Tag hinzufgen
        bisZeitPunkt.add(Calendar.DAY_OF_YEAR, 1);

        return bisZeitPunkt.getTimeInMillis();
    }

    /**
     * Ermittelt das IntervallEnde ausgehend von einem Zeitpunkt und der Intervalldauer Monat
     *
     * @param fruehestesIntervallEnde Endzeitpunkt, bis zu dem mindestens gerechnet werden muss.
     *
     * @return Intervallende (AUSSCHLIESSLICH!!), also der Intervallbeginn des nchsten Intervalls
     */
    private long ermittleIntervallEndeLZZS(long fruehestesIntervallEnde) {
        Calendar bisZeitPunkt = new GregorianCalendar();

        bisZeitPunkt.setTimeInMillis(ermittleIntervallBeginnLZZS(fruehestesIntervallEnde));

        // Zum Intervallanfang 1 Monat hinzufgen
        bisZeitPunkt.add(Calendar.MONTH, 1);

        return bisZeitPunkt.getTimeInMillis();
    }

    /**
     * Folgeauftrag fr den aktuell bergebenen Auftrag berechnen.
     * <p/>
     * Der Folgeauftrag liegt genau eine Intervalllnge spter.
     *
     * @param auftrag Der aktuelle Auftrag, fr den ein Folgeauftrag berechnet werden soll.
     *
     * @return Folgeauftrag fr den aktuell als Parameter bergebenen Auftrag oder <code>null</code>, wenn kein
     *         Folgeauftrag mehr existiert.
     */
    private Auftrag erstelleFolgeAuftrag(Auftrag auftrag) {
        return new Auftrag(auftrag);
    }

    /** Initialisierung von AggrLzzsDTV */
    private void initialisierung() {

        // VewLZZS-Objekte konstruieren
        // Parameter fr den Bast-Export wird in dieser Form erwartet !?
        String  ftpServer = String.format("Benutzer:%sPasswort:%sURL:%sPassivmodus:%s", _ftpServerBenutzer, _ftpServerPasswort, _ftpServerURL, _ftpPassivModus);
        VewLZZS vewLZZS   = new VewLZZS(_dav, _listeLzzsSystemObjekte, _verwaltungsObjektBastExportSteuerung, ftpServer, _isAenderungDatenstruktur, _isBastBandUeberschreiben);

        // ---------------------------------------------------------------------------------
        // Initiales setzen der Auftragsliste auf Basis der VewLZZS und der Aurufparameter
        // ---------------------------------------------------------------------------------
        long          ausfuehrungsZeitpunktMonat = _bisZeitPunktLZZS + (_minimaleWarteZeitInStundenLZZS * 60 * 60 * 1000);
        long          ausfuehrungsZeitpunktTag   = _bisZeitPunktAXL + (_minimaleWarteZeitInStundenAXL * 60 * 60 * 1000);
        AuftragsListe auftragsListe              = AuftragsListe.getInstanz();
        Auftrag       auftrag;

        // Auftrag fr die MONATS-Werte
        auftrag = new Auftrag(ausfuehrungsZeitpunktMonat, _vonZeitPunktLZZS, _bisZeitPunktLZZS, vewLZZS, BastFormat.V2004);
        auftragsListe.addAuftrag(auftrag);
        auftrag = new Auftrag(ausfuehrungsZeitpunktMonat, _vonZeitPunktLZZS, _bisZeitPunktLZZS, vewLZZS, BastFormat.V2007);
        auftragsListe.addAuftrag(auftrag);

        // Auftrag fr die Tages-Werte
        auftrag = new Auftrag(ausfuehrungsZeitpunktTag, _vonZeitPunktAXL, _bisZeitPunktAXL, vewLZZS, BastFormat.AXL2009);
        auftragsListe.addAuftrag(auftrag);
    }

    /**
     * Initialisierung. Wird durch den {@link de.bsvrz.sys.funclib.application.StandardApplicationRunner} aufgerufen.
     *
     * @param dav Die Datenverteilerverbindung.
     */
    @Override
    public void initialize(ClientDavInterface dav) {
        _dav           = dav;
        _konfiguration = _dav.getDataModel();
    }

    //~ GET METHODEN ==========================================================

    /**
     * Diese Methode muss von der Applikation implementiert werden, um zu signalisieren, ob Sendesteuerungen erwnscht
     * sind und mit der Methode <code>dataRequest</code> verarbeitet werden. In der Implementierung dieser Methode
     * drfen keine synchronen Aufrufe, die auf Telegramme vom Datenverteiler warten (wie z.B. Konfigurationsanfragen)
     * durchgefhrt werden, da ansonsten ein Deadlock entsteht.
     *
     * @param object          Das in der zugehrigen Sendeanmeldung angegebene System-Objekt.
     * @param dataDescription Die in der zugehrigen Sendeanmeldung angegebenen beschreibenden Informationen der
     *                        angemeldeten Daten.
     *
     * @return <code>true</code>, falls Sendesteuerungen gewnscht sind, sonst <code>false</code>.
     *
     * @see #dataRequest
     */
    @Override
    public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
        return false;
    }

    //~ METHODEN ==============================================================

    /**
     * Wertet die Aufrufparameter aus.
     *
     * @param argumentList Aufrufparamter der Applikation
     *
     * @throws Exception Wenn die Aufrufparameter nicht vollstndig ausgewertet werden knnen.
     */
    @Override
    public void parseArguments(ArgumentList argumentList) throws Exception {
        debug = Debug.getLogger();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -pidVerwaltungsObjekt
        // ------------------------------------------------------------------------------------------------------
        _pidVerwaltungsObjektBastExportSteuerung = argumentList.fetchArgument("-pidVerwaltungsObjekt=automatisch").asNonEmptyString();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -listePids
        // ------------------------------------------------------------------------------------------------------
        // _listePidsAufrufParameter = argumentList.fetchArgument("-listePids=dz.scheppau.a39.49_400").asNonEmptyString();
        //
        _listePidsAufrufParameter = argumentList.fetchArgument("-listePids=alle").asNonEmptyString();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunktLZZS
        // ------------------------------------------------------------------------------------------------------
        try {
            _vonZeitPunktAufrufParameterLZZS = argumentList.fetchArgument("-vonZeitPunktLZZS=jetzt").asTime();
        }
        catch (Exception e) {
            debug.error(String.format("Ungltiges Argument beim Aufrufparameter [-vonZeitPunktLZZS]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunktAXL
        // ------------------------------------------------------------------------------------------------------
        try {
            _vonZeitPunktAufrufParameterAXL = argumentList.fetchArgument("-vonZeitPunktAXL=jetzt").asTime();
        }
        catch (Exception e) {
            debug.error(String.format("Ungltiges Argument beim Aufrufparameter [-vonZeitPunktAXL]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInStundenLZZS
        // ------------------------------------------------------------------------------------------------------
        _minimaleWarteZeitInStundenAufrufParameterLZZS = argumentList.fetchArgument("-minimaleWarteZeitInStundenLZZS=1").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInStundenAXL
        // ------------------------------------------------------------------------------------------------------
        _minimaleWarteZeitInStundenAufrufParameterAXL = argumentList.fetchArgument("-minimaleWarteZeitInStundenAXL=1").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -bastBandUeberschreiben
        // ------------------------------------------------------------------------------------------------------
        try {
            _isBastBandUeberschreibenAufrufParameter = argumentList.fetchArgument("-bastBandUeberschreiben=ja").booleanValue();
        }
        catch (Exception e) {
            debug.error(String.format("Ungltiges Argument beim Aufrufparameter [-bastBandUeberschreiben]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -aenderungDatenstruktur
        // ------------------------------------------------------------------------------------------------------
        try {
            _isAenderungDatenstrukturAufrufParameter = argumentList.fetchArgument("-aenderungDatenstruktur=nein").booleanValue();
        }
        catch (Exception e) {
            debug.error(String.format("Ungltiges Argument beim Aufrufparameter [-aenderungDatenstruktur]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -ftpServerXXX
        // ------------------------------------------------------------------------------------------------------
        _ftpServerBenutzerAufrufParameter = argumentList.fetchArgument("-ftpServerBenutzer=SWE_BAST").asNonEmptyString();
        _ftpServerPasswortAufrufParameter = argumentList.fetchArgument("-ftpServerPasswort=geheim").asNonEmptyString();
        _ftpServerURLAufrufParameter      = argumentList.fetchArgument("-ftpServerURL=localhost").asNonEmptyString();

        try {
            _ftpPassivModusAufrufParameter = argumentList.fetchArgument("-ftpPassivModus=nein").asNonEmptyString().toLowerCase();
        }
        catch (Exception ignored) {}

        if (!_ftpPassivModusAufrufParameter.equals("ja")) {
            _ftpPassivModusAufrufParameter = "nein";
        }
    }

    // Startet die eigentliche Aggregation, wertet die Auftragsliste aus und aggregiert die Daten.
    private void run() {
        debug.info(String.format("%n=================================================================%n" + "Programm AutoExport BASt-Band gestartet...%n" + "=================================================================%n"));

        // Auftragsliste holen
        AuftragsListe auftragsListe = AuftragsListe.getInstanz();

        // Endlosschleife mit Prfung der Auftragsliste.
        // Sobald keinerlei Auftrge mehr vorhanden sind wird das Programm beendet
        while (auftragsListe.hatNochAuftraege()) {
            Auftrag auftrag;
            Auftrag folgeAuftrag;

            if ((auftrag = auftragsListe.holeNaechstenAuftrag()) != null) {
                StringBuffer statusMeldung = new StringBuffer();

                // Ggf. Folgeauftrag in Liste eintragen
                folgeAuftrag = erstelleFolgeAuftrag(auftrag);
                statusMeldung.append(String.format("Aktueller Auftrag: %s", auftrag));

                if (folgeAuftrag != null) {
                    auftragsListe.addAuftrag(folgeAuftrag);
                    statusMeldung.append(String.format("%nFolgeauftrag     : %s", folgeAuftrag));
                } else {
                    statusMeldung.append(String.format("%nFolgeauftrag     : %s", "kein weiterer Folgeauftrag mehr zu bercksichtigen"));
                }

                debug.info(statusMeldung.toString());

                // --------------------------
                // Auftrag ausfhren...
                // --------------------------
                auftrag.ausfuehren();
                debug.info(String.format("Ausgefhrter Auftrag: %s", auftrag));
            } else {

                // Zurzeit ist nichts zu tun, also warten und es dann noch mal versuchen...
                debug.info(String.format("Zur Zeit keine sofort auszufhrenden Auftrge in der Auftragsliste. Warte %d Sekunden.", WARTEZEIT_AUF_NAECHSTEN_AUFTRAG / 1000));

                try {
                    Thread.sleep(WARTEZEIT_AUF_NAECHSTEN_AUFTRAG);
                }
                catch (InterruptedException e) {
                    debug.info("Sleep der Auftragsbearbeitung wurde vorzeitig unterbrochen: ", e);
                }
            }
        }

        debug.info(String.format("%n=================================================================%n" + "Programm AutoExport BASt-Band wird ordnungsgem beendet,%n" + "da keinerlei Auftrge (weder aktuell noch zuknftig) anliegen%n" + "=================================================================%n"));
        System.exit(0);
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 31.10.12
